/* 
 * rlocated.c
 *
 * Copyright (C) 2004 Rasto Levrinc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */ 

/* rlocated is a daemon that reads rlocate dev file and copies it to the
 * rlocate db file every INTERVAL seconds.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "../pidfile.h"
#include <string.h>
#include <paths.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <getopt.h>
#include <grp.h>
#include <config.h>
#include <syslog.h>
#include <limits.h>
#include "../misc.h"

#ifdef HAVE_FTS_H
#include <fts.h>
#else
#include "sl_fts.h"
#endif

#define RL_VERSION "rlocate daemon " VERSION "\n"

#define INTERVAL 2
#define ROOT_UID 0

static char *OUTPUT          = NULL; // can be set from command line
static int  OUTPUT_OPTION    = 0; // 1 if --output option was specified
static char *MODULE_VERSION  = NULL; // is set in parse_proc
static char *EXCLUDE_DIR     = NULL; // is set in parse_proc
static int  COUNTDOWN        = 0; // used for testing, always set it to 0
static char *RLOCATEDEV      = "/dev/rlocate";
static char *RLOCATEPROC     = "/proc/rlocate";
static char *RLOCATE_DIFF_DB = NULL;


static char *PidFile = _PATH_VARRUN "rlocated.pid";
static int  NO_DAEMON = 0; /* don't run in daemon mode */
static int  NO_LOOP   = 0; /* don't run in loop */
static FILE *fd_dev;        /* file handle for dev file */
static FILE *fd_db;         /* file handle for db file  */
static int  fd_proc;       /* file handle for proc file  */
static int  fd_cfg;        /* file handle for module config file  */

static char buf[BUFSIZ];   /* buffer for copy from dev file to db file */
static int RELOAD_CONFIG;      /* if set to 1, the config will be reloaded */
static char *PROGNAME;

/*
 * print_log()
 */
void print_log(int level, char *format_str, ...) {
        va_list ap;
        va_start(ap, format_str);
        if (NO_DAEMON) {
                fprintf(stderr, "%s: ", PROGNAME);
                if (level == LOG_WARNING)
                        fprintf(stderr, "WARNING: ");
                vfprintf(stderr, format_str, ap);
                fputc('\n', stderr); fflush(stderr);
        } else {
                vsyslog(level, format_str, ap );
        }
        va_end(ap);
}

/*
 * clean_up
 */
static void clean_up()
{
        if (fd_db != 0) {
                fclose(fd_db);
        }
        if (fd_dev != 0) {
                fclose(fd_dev);
        }
        if (RLOCATE_DIFF_DB != NULL)
                free(RLOCATE_DIFF_DB);
        if (MODULE_VERSION != NULL)
                free(MODULE_VERSION);
        if (EXCLUDE_DIR != NULL)
                free(EXCLUDE_DIR);
        print_log(LOG_INFO, "rlocated daemon terminated");
        exit(1);
}

/*
 *  * xmalloc() allocate n bytes with malloc.
 *   */
void *xmalloc(const unsigned n) {
        void *p;
        p = malloc(n);
        if (p) return p;
        print_log(LOG_ERR, "malloc: %s\n", strerror(errno));
        clean_up();
        exit(1);
}

/*
 * xstrdup() return copy of string.
 */
void *xstrdup(const char *s) {
        void *p;
        p = strdup(s);
        if (p) return p;
        print_log(LOG_ERR, "strdup: %s\n", strerror(errno));
        clean_up();
        exit(1);
}


/*
 * stop_daemon
 */
void stop_daemon(sig)
        int sig;
{
        clean_up();
        return;
}

/*
 * reload_config
 */
void reload_config(sig)
        int sig;
{
        print_log(LOG_INFO, "reloading config");
        RELOAD_CONFIG = 1;
        return;
}
/* 
 * Usage 
 */

void usage() 
{
        printf("%s\n"
               "Copyright (c) 2004 Rasto Levrinc\n\n"
               "usage: %s [-n] [-o <file>] [--output=<file>]\n\n"
               "   Options:\n"
               "   -n --nodaemon        - Don't run in daemon mode.\n"
               "   -l --noloop          - Don't run in loop.\n"
               "   -o <file>\n"
               "   --output=<file>      - Specifies the rlocate database.\n"
               "   -h --help            - Display this help.\n"
               "   -V --version         - Display version.\n"
               "\n",
               RL_VERSION, PROGNAME);
               exit(0);
} 

/*
 * cfg_to_proc() is used, so that no error is reported twice. After
 * every error it sleeps for 60 seconds.
 */
int cfg_to_proc(const int type)
{
        static int last_error = 0;
        static int error = 0;
        error++;
        if (type == 0) 
                return 0;
        else if (type == -1) {
                error = 0;
                last_error = 0;
                return 0;
        }
        if (fd_cfg>0) {
                if (close(fd_cfg) < 0) 
                        print_log(LOG_WARNING, "cannot close %s: %s", 
                                  MODULE_CFG, strerror(errno) );
        }
        if (fd_proc>0) {
                if (close(fd_proc) < 0) 
                        print_log(LOG_WARNING, "cannot close %s: %s", 
                                  RLOCATEPROC, strerror(errno) );
        }
        if ( last_error == error) {
                error = 0;
                sleep(60);
                return 0;
        } else {
                last_error = error;
                error = 0;
                return 1;
        }
}

int cfg_to_proc_error(void)
{
        return cfg_to_proc(1);
}

void cfg_to_proc_ok(void)
{
        cfg_to_proc(0);
}

void cfg_to_proc_done(void)
        
{
        cfg_to_proc(-1);
}

/*
 * init_module() copies content from module config file to proc file and 
 * activate module. If something of that fails retry every minute. 
 */
void init_module(void) 
{
        struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, 0 };
        auto ssize_t n;

        while (1) {
                if ( (fd_cfg = open(MODULE_CFG, O_RDONLY)) < 0 ) {
                        if (cfg_to_proc_error())
                                print_log(LOG_WARNING, 
                                       "updatedb must be run first, waiting...");
                        continue; 
                } else
                        cfg_to_proc_ok();
                        
                if ( (fd_proc = open(RLOCATEPROC, O_WRONLY, 0)) < 0 ) {
                        if (cfg_to_proc_error()) 
                                print_log(LOG_WARNING, 
                                      "rlocate module is not loaded, waiting...");
                         
                        continue;
                } else 
                        cfg_to_proc_ok();

                fl.l_type = F_RDLCK;
                if (fcntl(fd_cfg, F_SETLKW, &fl) < 0) {
                        if (cfg_to_proc_error())
                                print_log(LOG_WARNING, 
                                          "fcntl lock error: %s", 
                                          strerror(errno) );
                        continue;
                } else 
                        cfg_to_proc_ok();
                
                /* activate module */
                if (write(fd_proc, "activated:1", 12) != 12) {
                        if (cfg_to_proc_error())
                                print_log(LOG_WARNING, "write error %s: %s", 
                                          RLOCATEPROC, strerror(errno) );
                        continue;
                } else
                        cfg_to_proc_ok();
                /* copy module cfg file to proc file. */
                while((n = read(fd_cfg, buf, BUFSIZ)) !=0 ) {
                        if (write(fd_proc, buf, n) != n) {
                                if (cfg_to_proc_error())
                                        print_log(LOG_WARNING, 
                                                  "write error %s: %s", 
                                                  RLOCATEPROC, strerror(errno) );
                        } else {
                                cfg_to_proc_ok();
                        }
                }
                break;
        }
        cfg_to_proc_done();
        /* release lock */
        fl.l_type = F_UNLCK;
        if (fcntl(fd_cfg, F_SETLKW, &fl) < 0)
                print_log(LOG_WARNING, "fcntl unlock error: %s", 
                          strerror(errno) );

        if (close(fd_cfg) < 0) 
                print_log(LOG_WARNING, "cannot close %s: %s", MODULE_CFG,
                                                              strerror(errno) );
        if (close(fd_proc) < 0)
                print_log(LOG_WARNING, "cannot close %s: %s", RLOCATEPROC, 
                                                              strerror(errno) );
}

/*
 * get_diff_db_name() returns a dbname with ".diff" attached.
 */
char *get_diff_db_name(const char *dbname)
{
        int db_len = strlen(dbname);      
        char *diff_dbname = xmalloc( db_len + 6 );
        strcpy(diff_dbname, dbname);
        strcpy(diff_dbname + db_len, ".diff");
        return diff_dbname;
}

/*
 * parse_proc() parse /proc/rlocate and set RLOCATE_DIFF_DB, MODULE_VERSION and
 * EXCLUDE_DIR global variables.
 */

void parse_proc()
{
        char *fbuf   = NULL;
        char *str    = NULL;
        char *endstr;
        fbuf = load_file(RLOCATEPROC);
        if (fbuf) {
                /* MODULE_VERSION */
                str = strstr(fbuf, "version: ");
                if (str) {
                        str+=9;
                        endstr = strstr(str, "\n");
                        *endstr = '\0';
                        /* MODULE_VERSION can be allocated by reload
                         * so it will be freed before it will be
                         * allocated again */
                        if (MODULE_VERSION != NULL)
                                free(MODULE_VERSION);
                        MODULE_VERSION = xstrdup( str );
                        *endstr = '\n';
                } else {
                        free(fbuf);
                        return;
                }
                str = endstr;

                /* EXCLUDE_DIR */
                str = strstr(fbuf, "\nexcludedir: ");
                if (str) {
                        str+=13;
                        endstr = strstr(str, "\n");
                        *endstr = '\0';
                        if (EXCLUDE_DIR != NULL)
                                free(EXCLUDE_DIR);
                        EXCLUDE_DIR = xstrdup( str );
                        *endstr = '\n';
                } else {
                        free(fbuf);
                        return;
                }
                str = endstr;

                /* RLOCATE_DIFF_DB */
                if (OUTPUT_OPTION == 0) { // --option has precedence
                        str = strstr(fbuf, "\noutput: ");
                        if (str) {
                                str+=9;
                                endstr = strstr(str, "\n");
                                *endstr = '\0';
                                if (RLOCATE_DIFF_DB != NULL)
                                        free(RLOCATE_DIFF_DB);
                                RLOCATE_DIFF_DB = get_diff_db_name(str);
                                *endstr = '\n';
                        }
                } else {
                        free(fbuf);
                        return;
                }
                free(fbuf);
        }
}

/* Check to see if a path matches an excluded one. */

int match_exclude(char *path, char *filedir) {
        int res=0;
        char *newstr;
   
        newstr = (char *)xmalloc(strlen(path)+strlen(filedir)+3);
        newstr[0] = '\0';
        strcat(newstr,"*");
        strcat(newstr,path);
        strcat(newstr,filedir);   
        strcat(newstr,"*");
    
        if (strstr(EXCLUDE_DIR,newstr) != NULL)       
                res=1;
    
        free(newstr);
 
        return res;
}

/*
 * traverse_dir() traverse directory 'dirstr' and put the paths, that match 
 * the pattern, to the tree.
 */
void traverse_dir(char *dirstr)
{
        FTSENT *file;
        FTS *dir;
        int i;
        char **dir_array;
        *dirstr = '/'; 
        int dirstr_len = strlen(dirstr);
        dirstr[dirstr_len-1] = '\0'; // remove '\n'
        dir_array = (char **)xmalloc(sizeof(char **)*2);
        *dir_array = dirstr;
        dir_array[1] = NULL;
        dir = fts_open(dir_array, FTS_PHYSICAL|FTS_NOSTAT, NULL);  
        for (i = 0; i > -1; i+=1) {
                file = fts_read(dir);
                
                if (!file)
                        break;
                
                if (file->fts_info != FTS_DP && file->fts_info != FTS_NS) {
                        if ((EXCLUDE_DIR!='\0' && 
                             !match_exclude(file->fts_path,"")) || 
                            EXCLUDE_DIR=='\0') {
                                // write to the db without leading '/', with
                                // '\n' at the end of the line.
                                strcpy(dirstr, file->fts_path+1);
                                dirstr_len = strlen(dirstr);
                                dirstr[dirstr_len]='\n';
                                dirstr[dirstr_len+1]='\0';
                                if (fputs(dirstr, fd_db) == EOF)
                                        print_log(LOG_WARNING, 
                                                  "write error %s: %s", 
                                                  RLOCATE_DIFF_DB, 
                                                  strerror(errno) );
                        } else {
                                fts_set(dir,file,FTS_SKIP);
                        }
                }
        }
        fts_close(dir);
        free(dir_array);
}

/*
 * main
 */
int main(int argc, char **argv) 
{
        int ch; /* option character */
        char *p;
        int this_option_optind;
        int option_index;
        struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, 0 };
        int no_module; // 0, if kernel module is not loaded. 1, if it is.
        int num_fds;
        auto int fd;
        auto int fd_db_no;
        char buffer[PATH_MAX+1];

        struct group *grpres;

        struct option long_options[] =
        {
                {"help",    0, 0, 'h'},
                {"nodaemon",0, 0, 'n'},
                {"version", 0, 0, 'V'},
                {"output",  1, 0, 'o'},
                {"noloop",  0, 0, 'l'},
                {0, 0, 0, 0}
        };
        uid_t uid = getuid();

        PROGNAME = ((p = strrchr(argv[0], '/')) ? p+1 : *argv);
        /* Parse the command-line */
        while (1) {
                this_option_optind = optind ? optind : 1;
                option_index = 0;
                ch = getopt_long (argc, argv, "hVno:l",
                                  long_options, &option_index);
                if (ch == -1)
                        break;
                switch(ch) {
                        case 'h':
                                usage();
                                break;
                        case 'n':       /* don't run in daemon mode */
                                NO_DAEMON++;
                                break;
                        case 'l':       
                                NO_LOOP++;
                                NO_DAEMON++;
                                break;
                        case 'V':
                                printf("%s", RL_VERSION);
                                exit (1);
                                break;
                        case 'o':
                                if (optarg && strlen(optarg)>0) {
                                        OUTPUT_OPTION = 1;
                                        RLOCATE_DIFF_DB=get_diff_db_name(optarg);
                                }
                                break;
                        default:
                                return(1);
                                break;
                }
        }
        /* open log */
        if (!NO_DAEMON)
                openlog(PROGNAME, LOG_PID, LOG_DAEMON);

        /* set gid and check if rlocate group exists */
        if ((grpres = getgrnam(RLOCATE_GRP)) == NULL) {
                print_log(LOG_WARNING, "Could not find the group: %s in the "
                                       "/etc/group file.", RLOCATE_GRP);
                print_log(LOG_WARNING, "This is a result of the group missing "
                                       "or a corrupted group file.");
        } else {
                setgid(grpres->gr_gid);
        }

        if (uid != ROOT_UID) {
                print_log(LOG_ERR, "you are not authorized to run this program");
                exit(1);
        }

        /* check the pid file and daemonize */
        if (!NO_DAEMON) {
                if (!check_pid(PidFile)) {
                        if ( fork() == 0 ) {
                                num_fds = getdtablesize();
                                for (fd = 0; fd < num_fds; ++fd) {
                                        close(fd);
                                }
                                setsid();
                        } else 
                                exit(0);
                } else {
                        print_log(LOG_ERR, "already running, exiting...");
                        exit(1);
                }
        }
        if (!NO_LOOP) {
                if (!check_pid(PidFile)) {
                        if (!write_pid(PidFile))
                                clean_up();
                } else {
                        print_log(LOG_ERR, "already running, exiting...");
                        clean_up();
                }
                print_log(LOG_INFO, "rlocated daemon started");
                /* Signals */
                for (ch= 1; ch < NSIG; ++ch)
                        signal(ch, SIG_IGN);
                signal(SIGINT, stop_daemon);
                signal(SIGKILL, stop_daemon);
                signal(SIGTERM, stop_daemon);
                signal(SIGHUP, reload_config);
        }

        fl.l_pid = getpid();
        RELOAD_CONFIG = 1;
        /* the main loop */
        no_module = 1;
        while (1) {
                if (no_module && !NO_LOOP) {
                        init_module();
                        print_log(LOG_INFO, "rlocate module is initialized");
                }
                no_module = 0;

                if (RELOAD_CONFIG>0) {
                        parse_proc();
                        // compare daemon and module versions
                        if (strcmp(MODULE_VERSION, VERSION)) {
                                print_log(LOG_WARNING, "module version (%s) "
                                          "doesn't match daemon version (%s)", 
                                          MODULE_VERSION, VERSION);
                        }
                }

                RELOAD_CONFIG = 0;
                if (!NO_LOOP)
                        sleep(INTERVAL);
                /* open db file */
                if ( (fd_db = fopen(RLOCATE_DIFF_DB, "a")) == NULL ) {
                        print_log(LOG_WARNING, "cannot open %s for writing: %s", 
                                  RLOCATE_DIFF_DB, strerror(errno));
                        if (NO_LOOP) clean_up();
                        continue;
                }
                fl.l_type = F_WRLCK;
                fd_db_no = fileno(fd_db);
                if (fcntl(fd_db_no, F_SETLKW, &fl) < 0) {
                        print_log(LOG_WARNING, "fcntl lock error: %s", 
                                  strerror(errno) );
                        RELOAD_CONFIG = 1;
                        if (NO_LOOP) clean_up();
                        continue;
                }

                /* open dev file */
                if ( (fd_dev = fopen(RLOCATEDEV, "r")) == NULL ) {
                        print_log(LOG_WARNING, "cannot open %s: %s", 
                                  RLOCATEDEV, strerror(errno));
                        no_module = 1;
                        if (NO_LOOP) clean_up();
                        continue;
                }
                /* copy all lines from dev file to the diff database. The
                 * lines that start with 'm' are directories, that where 
                 * moved, so they are traversed and written to the diff 
                 * database.
                 */
                while ( (fgets(buffer, PATH_MAX+1, fd_dev)) != NULL ) {
                        if (*buffer == 'm') {
                                traverse_dir(buffer);
                        } else {
                                // write without leading '/'
                                if (fputs(buffer+1, fd_db) == EOF)
                                        print_log(LOG_WARNING, 
                                                  "write error %s: %s", 
                                                  RLOCATE_DIFF_DB, 
                                                  strerror(errno) );
                        }
                }
                if (ferror(fd_dev))
                        print_log(LOG_WARNING, "read error %s: %s", 
                                  RLOCATEDEV, strerror(errno) );


                if (fclose(fd_dev) < 0)
                        print_log(LOG_WARNING, "cannot close %s: %s", 
                                  RLOCATEDEV, strerror(errno) );
                /* release lock */
                fl.l_type = F_UNLCK;

                if (fcntl(fd_db_no, F_SETLKW, &fl) < 0)
                        print_log(LOG_WARNING, "fcntl unlock error: %s", 
                                  strerror(errno) );
                if (fclose(fd_db) < 0) 
                        print_log(LOG_WARNING, "cannot close %s: %s", 
                                  RLOCATE_DIFF_DB, strerror(errno) );
                if (NO_LOOP) break;

                if (COUNTDOWN != 0 ) {
                        if (COUNTDOWN == 1)
                                NO_LOOP = 1;
                        COUNTDOWN--;
                }
        }
        if (RLOCATE_DIFF_DB != NULL)
                free(RLOCATE_DIFF_DB);
        if (OUTPUT != NULL)
                free(OUTPUT);
        if (MODULE_VERSION != NULL)
                free(MODULE_VERSION);
        if (EXCLUDE_DIR != NULL)
                free(EXCLUDE_DIR);
        return 0;
}
