/*****************************************************************************
 * Real-Time Locate
 *
 * Copyright (c) 2004 Rasto Levrinc
 *
 * Real-Time Locate: http://stud3.tuwien.ac.at/~e9526925/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <string.h>
#include <fnmatch.h>
#include <regex.h>
#define __USE_GNU
#include <search.h>
#include <limits.h>
#include <ctype.h>
#include <signal.h>
#include <paths.h>

#include "misc.h"
#include "pidfile.h"

#ifdef HAVE_FTS_H
#include <fts.h>
#else
#include "sl_fts.h"
#endif

/* GLOBALS */
#define WARNING 0
#define FATAL 1

#define RLOCATEPROC  "/proc/rlocate"

#define SLOC_UID 0

/* extern variables are defined in main.c */
extern char *progname;
extern int QUIET;
extern uid_t UID;
extern int max_queries;
extern int NOCASE;
extern int REGEXP;
extern int VERBOSE;
extern char *EXCLUDE;
extern char *EXCLUDE_DIR;
extern int INITDIFFDB; /* initialize diff database if it's set to 2 */

extern int check_path_access(char *codedpath);
extern int match_exclude(char *path, char *filedir);

/* global variables STR, CASESTR, GLOBFLAG and PREG are set in rlocate_init()*/
static char *STR; 
static char *CASESTR;
static int GLOBFLAG;
static regex_t *PREG;

static char *PidFile = _PATH_VARRUN "rlocated.pid";
static int rlocate_queries; // number of results

static void *paths_tree_root;          // root of the tree, that contains paths
static char *tmp_output_diff = NULL;   // temp output diff database

typedef struct paths_list { // list, that contains paths that where added to  
        char *path;         // the filesystem
        struct paths_list *next;
} Paths_list;

Paths_list *paths_list_current; // pointer to the current path
Paths_list *paths_list_root;    // pointer to the root of the list of paths

/*
 * xmalloc() allocate n bytes with malloc and exit if there is an error.
 */
void *xmalloc(const unsigned n) {
        void *p;
        p = malloc(n);
        if (p) return p;
        report_error(FATAL, QUIET, "%s: malloc: %s\n",
                                     progname, strerror(errno));
        exit(1);
}

/*
 * xstrdup() return copy of string and exit if there is an error.
 */
void *xstrdup(const char *s) {
        void *p;
        p = strdup(s);
        if (p) return p;
        report_error(FATAL, QUIET, "%s: strdup: %s\n",
                                     progname, strerror(errno));
        exit(1);
}
/*
 * write_to_fd writes() string to the fd file descriptor. fd must be open.
 */
void write_to_fd(const int fd, const char *filename, const char *string ) {
        int len = strlen(string);
        if (write(fd, string, len) != len)
                report_error(WARNING, QUIET, "%s: write error %s: %s\n",
                                        progname, filename, strerror(errno) );
}

/*
 * get_diff_db_name() returns a dbname with ".diff" attached.
 */
char* get_diff_db_name(const char *dbname)
{
        int db_len = strlen(dbname);
        char *diff_dbname = (char *)xmalloc( db_len + 6 );
        strcpy(diff_dbname, dbname);
        strcpy(diff_dbname + db_len, ".diff");
        return diff_dbname;
}

/*
 * get_tmp_db_name() returns a dbname with ".tmp" attached.
 */
char *get_tmp_db_name(const char *dbname)
{
        int db_len = strlen(dbname);
        char *tmp_dbname = (char *)xmalloc( db_len + 5 );
        strcpy(tmp_dbname, dbname);
        strcpy(tmp_dbname + db_len, ".tmp");
        return tmp_dbname;
}

/*
 * run_rlocated() run rlocated with --noloop option if user is root.
 */
void run_rlocated()
{
        pid_t pid;

        if (UID == SLOC_UID) {
                if ( (pid = fork()) < 0 )
                        report_error(WARNING, QUIET, 
                                "%s: rlocate_init(): fork error\n", progname);
                else if (pid == 0) {
                        if (execlp(RLOCATED_CMD, "rlocated", "--noloop", NULL)<0) 
                                report_error(WARNING, QUIET,
                                        "%s: rlocate_init(): execlp error\n", 
                                        progname);
                }
                if (waitpid(pid, NULL, 0) < 0)
                        report_error(WARNING, QUIET,
                                "%s: rlocate_init(): waitpid error\n",
                                progname);
        }
}

/*
 * rlocate_ftscompare()
 */
int rlocate_ftscompare(const FTSENT **e1, const FTSENT **e2) {
        return strcmp((*e1)->fts_accpath, (*e2)->fts_accpath);
}

/*
 * rlocate_start_updatedb() is called before original updatedb creates its 
 * database. It moves the rlocate diff database to temp rlocate diff db.
 */
void rlocate_start_updatedb(const char *output, const char *startingpath) {
        int fd_cfg;
        int fd_db;
        char *output_diff;
        pid_t daemon_pid;
        struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, 0 };
        tmp_output_diff = NULL;
        // do it only for user root and if --initdiffdb option is specified 
        if (UID != SLOC_UID || INITDIFFDB == 0) 
                return;

        /* generate module.cfg file */
        if (UID == SLOC_UID && INITDIFFDB == 1) {
                if ( (fd_cfg=open(MODULE_CFG,O_WRONLY|O_CREAT|O_TRUNC, 0))<0 ) {
                        report_error(WARNING, QUIET, 
                             "%s: rlocate_start_updatedb(): open: could not "
                             "open %s: %s\n", 
                             progname, MODULE_CFG, strerror(errno));
                } else {
                        write_to_fd(fd_cfg, MODULE_CFG, 
                            "# This file is automatically "
                            "generated -- please do not edit.\nexcludedir:");
                        if (EXCLUDE_DIR!=NULL)
                                write_to_fd(fd_cfg, MODULE_CFG, EXCLUDE_DIR);
                        write_to_fd(fd_cfg, MODULE_CFG, "\nstartingpath:");
                        if (startingpath)
                                write_to_fd(fd_cfg, MODULE_CFG, startingpath);
                        write_to_fd(fd_cfg, MODULE_CFG, "\noutput:");
                        if (output)
                                write_to_fd(fd_cfg, MODULE_CFG, output);
                        if ( close(fd_cfg) < 0 )
                                report_error(WARNING, QUIET, 
                                     "%s: rlocate_start_updatedb(): close: "
                                     "can't close %s: %s\n", 
                                     MODULE_CFG, progname, strerror(errno) );
                }
	        /* update module /proc info */
                if ( (fd_db = open(RLOCATEPROC, O_WRONLY, 0)) < 0 ) {
                        report_error(WARNING, QUIET, 
                                        "%s: module not loaded? :%s\n",
                                        progname, strerror(errno));
                 return;
                }
                if (EXCLUDE_DIR != NULL) {
                        write_to_fd(fd_db, RLOCATEPROC, "excludedir:");
                        write_to_fd(fd_db, RLOCATEPROC, EXCLUDE_DIR);
                }
                if (startingpath != NULL) {
                        write_to_fd(fd_db, RLOCATEPROC, "\nstartingpath:");
                        write_to_fd(fd_db, RLOCATEPROC, startingpath);
                }
                if (startingpath != NULL) {
                        write_to_fd(fd_db, RLOCATEPROC, "\noutput:");
                        write_to_fd(fd_db, RLOCATEPROC, output);
                }
                write_to_fd(fd_db, RLOCATEPROC, "\nactivated: 1");

                if ( close(fd_db) < 0 )
                        report_error(WARNING, QUIET, 
                                "%s: rlocate_start_updatedb(): close: can't "
                                "close /proc/rlocate: %s\n", 
		                progname, strerror(errno) );

                /* reload rlocated daemon */
                daemon_pid = read_pid(PidFile);
                if (daemon_pid>0) {
                        kill(daemon_pid, 1);
                }

        } // UID == SLOC_UID

        /*rename rlocate diff db to temp rlocate diff db. If it's locked wait.*/
        output_diff = get_diff_db_name(output);

        if ( (fd_db = open(output_diff, O_WRONLY)) < 0 ) {
	        if (VERBOSE)
		        report_error(WARNING, QUIET ,
			        "%s: rlocate_start_updatedb(): open: could not "
			        "open rlocate diff database: %s: %s\n", 
				progname, output_diff, strerror(errno));
	} else {
                fl.l_pid = getpid();
                if (fcntl(fd_db, F_SETLKW, &fl) < 0) {
	                report_error(WARNING, QUIET, "%s: "
				     "rlocate_start_updatedb(): fcntl: can't "
				     "get a lock on %s: %s\n", 
					progname, output_diff, strerror(errno) );
                } else {
                        tmp_output_diff = get_tmp_db_name(output_diff);
	                /* rename releases the lock */
	                if ( rename(output_diff, tmp_output_diff) < 0 ) {
	                        report_error(WARNING, QUIET, "%s: "
					     "rlocate_start_updatedb(): rename:"
					     " can't rename to "
					     "%s: %s\n", 
						progname, tmp_output_diff, 
                                                strerror(errno) );
                                free(tmp_output_diff);
	                        tmp_output_diff = NULL;
                        }
                }
                if ( close(fd_db) < 0 )
                        report_error(WARNING, QUIET, "%s: "
				     "rlocate_start_updatedb(): close: can't "
				     "close %s: %s\n", 
						progname, output_diff,
                                                strerror(errno) );
        }
        free(output_diff);
}

/*
 * rlocate_end_updatedb() is called after original updatedb creates its 
 * diff database. It removes the temp rlocate diff database.
 */
void rlocate_end_updatedb() {
        /* remove temp rlocate diff database if it was created */
        if (tmp_output_diff != NULL) {
                if (unlink(tmp_output_diff) < 0)
                        report_error(WARNING, QUIET, 
                                     "%s: rlocate_end_updatedb(): unlink: "
                                     "can't unlink %s: %s\n", 
				     progname, tmp_output_diff, strerror(errno) );
                free(tmp_output_diff);
        }
}

/*
 * string_compare() compares two strings
 */
int string_compare(const void *string1, const void *string2)
{
        return strcmp((const char *) string1,
                      (const char *) string2);
}

/*
 * make_path() make copy of the path and add leading '/'.
 */
char *make_path(const char *path) {
        char *pathcopy;
        pathcopy = (char *)xmalloc(strlen(path)+2);
        strcpy(pathcopy+1, path);
        *pathcopy = '/';
        return pathcopy;
}
/*
 * check_path() returns 1 if the path matches the pattern. Path is
 * without leading '/'
 */
int check_path(const char *path)
{
        int foundit = 0;
        char *casecodedpath = NULL;
        int nmatch = 32;
        regmatch_t pmatch[32];
        char *cp = NULL;
        char *codedpath;
        codedpath = make_path(path);
	if (REGEXP) {
	        foundit = !regexec(PREG,codedpath,nmatch,pmatch,0);
	} else if (NOCASE) {
#ifdef FNM_CASEFOLD /* i suppose i also have strcasestr */
	        if (GLOBFLAG)
		        foundit =! fnmatch(STR,codedpath,FNM_CASEFOLD);
		else
		        foundit = (strcasestr(codedpath,STR) != NULL);
#else /* FNM_CASEFOLD */
                casecodedpath = xstrdup(codedpath);

		for (cp = casecodedpath; *cp; cp++)
		        *cp = tolower(*cp);
			
		if (GLOBFLAG)
		        foundit =! fnmatch(CASESTR,casecodedpath,0);
		else
		        foundit = (strstr(casecodedpath,CASESTR) != NULL);

		free(casecodedpath);
#endif /* FNM_CASEFOLD */
			
	} else {
		if (GLOBFLAG)
	                foundit =! fnmatch(STR,codedpath,0);
		else
		        foundit=(strstr(codedpath,STR) != NULL);			
        }
        free(codedpath);
        return foundit;
}

/* 
 * print_path() checks if path is accesible and prints it, if it is.
 */
void print_path(const char *path)
{
        char *pathcopy;
        pathcopy = make_path(path);
        /* pathcopy is freed in check_path_access function */
	if ( check_path_access(pathcopy)) { 
        	printf("/%s\n", path);
		rlocate_queries++;
	}
}

/* 
 * create_paths_list() is called from twalk and it creates the list of
 * paths, that where added to the filesystem.
 */
void create_paths_list(void *node, VISIT order, int level) 
{
        Paths_list *f;
        if (order == postorder || order == leaf) {
                f = (Paths_list*)xmalloc(sizeof(Paths_list));
                f->path = *(char **)node;
                f->next = NULL;
                if (paths_list_current == NULL)
                        paths_list_root = f;
                else 
                        paths_list_current->next = f;
                paths_list_current = f;
        }
}


/*
 * store_path() adds the file names that where added to the filesystem to the
 * tree, if they match the pattern.
 */
void store_path(const char *path, const char *str, const char *casestr, 
	        const int globflag, const regex_t *preg)
{
        char *pathcopy;
        void *node;
        if (! check_path(path))
                return;
        /* put the path to the tree of added file names. */
        pathcopy = xstrdup(path);
        node = tsearch((void *)pathcopy, &paths_tree_root, 
                               string_compare);
        if (node == NULL) {
                report_error(FATAL, QUIET, 
                                "%s: tsearch: insufficient memory\n", progname);
        }
        // free pathcopy if the path was in the tree before
        if (*(char**)node != pathcopy) {
                free(pathcopy);
        }
}

/*
 * rlocate_init() is called from original locate. It reads both rlocate and
 * temp rlocate diff databases and creates the list of paths.
 *  
 */
void rlocate_init(const char *rlocate_db, char *str, char *casestr, 
                  const int globflag, regex_t *preg)
{
        FILE *fd;
        char buffer[PATH_MAX+1];
        char *rlocate_diff_db;
        char *tmp_rlocate_diff_db;
        STR      = str;
        CASESTR  = casestr;
        GLOBFLAG = globflag;
        PREG     = preg;

        rlocate_diff_db     = get_diff_db_name(rlocate_db);
        tmp_rlocate_diff_db = get_tmp_db_name(rlocate_diff_db);

	rlocate_queries     = 0;
	paths_tree_root     = NULL;
	paths_list_current  = NULL; 
	paths_list_root     = NULL; 

        /* start rlocated once if the user is root, so that the database is 
         * up-to-date */
        run_rlocated();
        /* open and read rlocate diff database */
        if ( (fd = fopen(tmp_rlocate_diff_db, "r")) != NULL ) {
                while ( (fgets(buffer, PATH_MAX+1, fd)) != NULL ) {
                        buffer[strlen(buffer)-1]='\0';
                        store_path(buffer, str, casestr, globflag, preg);
                }
                if (fclose(fd) <0)
                        report_error(WARNING, QUIET, 
				     "%s: rlocate_init(): fclose: can't close "
				     "%s: %s\n", progname, tmp_rlocate_diff_db,
							   strerror(errno) );
        }
        if ( (fd = fopen(rlocate_diff_db, "r")) != NULL ) {
                while ( (fgets(buffer, PATH_MAX+1, fd)) != NULL ) {
                        buffer[strlen(buffer)-1]='\0';
                        store_path(buffer,str,casestr,globflag,preg);
                }
                if (fclose(fd) <0)
                        report_error(WARNING, QUIET, 
				     "%s: rlocate_init(): fclose: can't close "
				     "%s: %s\n", progname, rlocate_diff_db,
						        strerror(errno) );
        }
        free(rlocate_diff_db);
        free(tmp_rlocate_diff_db);
        twalk(paths_tree_root, (void*)create_paths_list); 
}

/*
 * rlocate_printit() is called from original locate, when it wants to print
 * the path that was found in its database. It returns amount of results.
 */
int rlocate_printit(const char *codedpath) 
{
        Paths_list *f;
        int str_ret;
	if (max_queries > 0 && rlocate_queries == max_queries) {
                return rlocate_queries;
        }
	/* print all paths from the list, that are alphabetically before the 
         * codedpath */
        while (paths_list_root != NULL && 
               (str_ret = strcmp(paths_list_root->path+1, codedpath+1)) <=0) {
                print_path(paths_list_root->path);
		if (max_queries > 0 && rlocate_queries == max_queries) {
                        return rlocate_queries;
                }
                f = paths_list_root;
                paths_list_root = paths_list_root->next;
                free(f); /* the path string will be freed, when the tree is 
                            destroyed */
        }
	/* print coded path, if it is not in the tree of added paths */
        if (tfind((void *)codedpath+1, &paths_tree_root, 
            string_compare) == NULL) { // ignore leading '/' in codedpath 
	        print_path(codedpath+1);
        }
	return rlocate_queries;
}

/*
 * free_string() is called from tdestroy
 */
void free_string(void *path)
{
        free((char **)path);
}

/*
 * rlocate_done() is called from original locate. It prints the rest of the 
 * paths in the list and cleans up the memory. It returns amount of results.
 */
int rlocate_done()
{
        Paths_list *f;
        // print the rest of the paths
        while ( paths_list_root != NULL ) {
		if (max_queries == 0 || rlocate_queries < max_queries)
                        print_path(paths_list_root->path);
                f = paths_list_root;
                paths_list_root = paths_list_root->next;
                free(f);
        }
        tdestroy(paths_tree_root, free_string);
	return rlocate_queries;
}
